# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the classes regarding a specific commit, such as
Commit, Modification,
ModificationType and Method.
"""
import logging
from _datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List, Set, Dict

import lizard
from git import Diff, Git, Commit as GitCommit, NULL_TREE

from pydriller.domain.developer import Developer

logger = logging.getLogger(__name__)


class ModificationType(Enum):
    """
    Type of Modification. Can be ADD, COPY, RENAME, DELETE, MODIFY or UNKNOWN.
    """

    ADD = 1
    COPY = 2
    RENAME = 3
    DELETE = 4
    MODIFY = 5
    UNKNOWN = 6


class Method:  # pylint: disable=R0902
    """
    This class represents a method in a class. Contains various information
    extracted through Lizard.
    """

    def __init__(self, func):
        """
        Initialize a method object. This is calculated using Lizard: it parses
        the source code of all the modifications in a commit, extracting
        information of the methods contained in the file (if the file is a
        source code written in one of the supported programming languages).
        """

        self.name = func.name
        self.long_name = func.long_name
        self.filename = func.filename
        self.nloc = func.nloc
        self.complexity = func.cyclomatic_complexity
        self.token_count = func.token_count
        self.parameters = func.parameters
        self.start_line = func.start_line
        self.end_line = func.end_line
        self.fan_in = func.fan_in
        self.fan_out = func.fan_out
        self.general_fan_out = func.general_fan_out
        self.length = func.length
        self.top_nesting_level = func.top_nesting_level


class Modification:  # pylint: disable=R0902
    """
    This class contains information regarding a modified file in a commit.
    """

    def __init__(self, old_path: str, new_path: str,
                 change_type: ModificationType,
                 diff_and_sc: Dict[str, str]):
        """
        Initialize a modification. A modification carries on information
        regarding the changed file. Normally, you shouldn't initialize a new
        one.
        """
        self._old_path = Path(old_path) if old_path is not None else None
        self._new_path = Path(new_path) if new_path is not None else None
        self.change_type = change_type
        self.diff = diff_and_sc['diff']
        self.source_code = diff_and_sc['source_code']
        self.source_code_before = diff_and_sc['source_code_before']

        self._nloc = None
        self._complexity = None
        self._token_count = None
        self._function_list = []

    @property
    def added(self) -> int:
        """
        Return the total number of added lines in the file.

        :return: int lines_added
        """
        added = 0
        for line in self.diff.replace('\r', '').split("\n"):
            if line.startswith('+') and not line.startswith('+++'):
                added += 1
        return added

    @property
    def removed(self):
        """
        Return the total number of deleted lines in the file.

        :return: int lines_deleted
        """
        removed = 0
        for line in self.diff.replace('\r', '').split("\n"):
            if line.startswith('-') and not line.startswith('---'):
                removed += 1
        return removed

    @property
    def old_path(self):
        """
        Old path of the file. Can be None if the file is added.

        :return: str old_path
        """
        if self._old_path is not None:
            return str(self._old_path)
        return None

    @property
    def new_path(self):
        """
        New path of the file. Can be None if the file is deleted.

        :return: str new_path
        """
        if self._new_path is not None:
            return str(self._new_path)
        return None

    @property
    def filename(self) -> str:
        """
        Return the filename. Given a path-like-string (e.g.
        "/Users/dspadini/pydriller/myfile.py") returns only the filename
        (e.g. "myfile.py")

        :return: str filename
        """
        if self._new_path is not None and str(self._new_path) != "/dev/null":
            path = self._new_path
        else:
            path = self._old_path

        return path.name

    @property
    def nloc(self) -> int:
        """
        Calculate the LOC of the file.

        :return: LOC of the file
        """
        self._calculate_metrics()
        return self._nloc

    @property
    def complexity(self) -> int:
        """
        Calculate the Cyclomatic Complexity of the file.

        :return: Cyclomatic Complexity of the file
        """
        self._calculate_metrics()
        return self._complexity

    @property
    def token_count(self) -> int:
        """
        Calculate the token count of functions.

        :return: token count
        """
        self._calculate_metrics()
        return self._token_count

    @property
    def methods(self) -> List[Method]:
        """
        Return the list of methods in the file. Every method
        contains various information like complexity, loc, name,
        number of parameters, etc.

        :return: list of methods
        """
        self._calculate_metrics()
        return self._function_list

    def _calculate_metrics(self):
        if self.source_code and self._nloc is None:
            analysis = lizard.analyze_file.analyze_source_code(self.filename,
                                                               self.source_code
                                                               )

            self._nloc = analysis.nloc
            self._complexity = analysis.CCN
            self._token_count = analysis.token_count

            for func in analysis.function_list:
                self._function_list.append(Method(func))

    def __eq__(self, other):
        if not isinstance(other, Modification):
            return NotImplemented
        if self is other:
            return True
        return self.__dict__ == other.__dict__


class Commit:
    """
    Class representing a Commit. Contains all the important information such
    as hash, author, dates, and modified files.
    """

    def __init__(self, commit: GitCommit, conf) -> None:
        """
        Create a commit object.

        :param commit: GitPython Commit object
        :param conf: Configuration class
        """
        self._c_object = commit

        self._modifications = None
        self._branches = None
        self._conf = conf

    @property
    def hash(self) -> str:
        """
        Return the SHA of the commit.

        :return: str hash
        """
        return self._c_object.hexsha

    @property
    def author(self) -> Developer:
        """
        Return the author of the commit as a Developer object.

        :return: author
        """
        return Developer(self._c_object.author.name,
                         self._c_object.author.email)

    @property
    def committer(self) -> Developer:
        """
        Return the committer of the commit as a Developer object.

        :return: committer
        """
        return Developer(self._c_object.committer.name,
                         self._c_object.committer.email)

    @property
    def project_name(self) -> str:
        """
        Return the project name.

        :return: project name
        """
        return Path(self._conf.get('path_to_repo')).name

    @property
    def author_date(self) -> datetime:
        """
        Return the authored datetime.

        :return: datetime author_datetime
        """
        return self._c_object.authored_datetime

    @property
    def committer_date(self) -> datetime:
        """
        Return the committed datetime.

        :return: datetime committer_datetime
        """
        return self._c_object.committed_datetime

    @property
    def author_timezone(self) -> int:
        """
        Author timezone expressed in seconds from epoch.

        :return: int timezone
        """
        return self._c_object.author_tz_offset

    @property
    def committer_timezone(self) -> int:
        """
        Author timezone expressed in seconds from epoch.

        :return: int timezone
        """
        return self._c_object.committer_tz_offset

    @property
    def msg(self) -> str:
        """
        Return commit message.

        :return: str commit_message
        """
        return self._c_object.message.strip()

    @property
    def parents(self) -> List[str]:
        """
        Return the list of parents SHAs.

        :return: List[str] parents
        """
        parents = []
        for p in self._c_object.parents:
            parents.append(p.hexsha)
        return parents

    @property
    def merge(self) -> bool:
        """
        Return True if the commit is a merge, False otherwise.

        :return: bool merge
        """
        return len(self._c_object.parents) > 1

    @property
    def modifications(self) -> List[Modification]:
        """
        Return a list of modified files.

        :return: List[Modification] modifications
        """
        if self._modifications is None:
            self._modifications = self._get_modifications()

        return self._modifications

    def _get_modifications(self):
        options = {}
        if self._conf.get('histogram'):
            options['histogram'] = True

        if self._conf.get('skip_whitespaces'):
            options['w'] = True

        if len(self.parents) == 1:
            # the commit has a parent
            diff_index = self._c_object.parents[0].diff(self._c_object,
                                                        create_patch=True,
                                                        **options)
        elif len(self.parents) > 1:
            # if it's a merge commit, the modified files of the commit are the
            # conflicts. This because if the file is not in conflict,
            # pydriller will visit the modification in one of the previous
            # commits. However, parsing the output of a combined diff (that
            # returns the list of conflicts) is challenging: so, right now,
            # I will return an empty array, in the meanwhile I will try to
            # find a way to parse the output.
            # c_git = Git(str(self.project_path))
            # d = c_git.diff_tree("--cc", commit.hexsha, '-r', '--abbrev=40',
            #                     '--full-index', '-M', '-p', '--no-color')
            diff_index = []
        else:
            # this is the first commit of the repo. Comparing it with git
            # NULL TREE
            diff_index = self._c_object.diff(NULL_TREE,
                                             create_patch=True,
                                             **options)

        return self._parse_diff(diff_index)

    def _parse_diff(self, diff_index) -> List[Modification]:
        modifications_list = []
        for diff in diff_index:
            old_path = diff.a_path
            new_path = diff.b_path
            change_type = self._from_change_to_modification_type(diff)

            diff_and_sc = {
                'diff': self._get_decoded_str(diff.diff),
                'source_code_before': self._get_decoded_sc_str(
                    diff.a_blob),
                'source_code': self._get_decoded_sc_str(
                    diff.b_blob)
            }

            modifications_list.append(Modification(old_path, new_path,
                                                   change_type, diff_and_sc))

        return modifications_list

    def _get_decoded_str(self, diff):
        try:
            return diff.decode('utf-8', 'ignore')
        except (UnicodeDecodeError, AttributeError, ValueError):
            logger.debug('Could not load the diff of a '
                         'file in commit %s', self._c_object.hexsha)
            return None

    def _get_decoded_sc_str(self, diff):
        try:
            return diff.data_stream.read().decode('utf-8', 'ignore')
        except (UnicodeDecodeError, AttributeError, ValueError):
            logger.debug('Could not load source code of a '
                         'file in commit %s', self._c_object.hexsha)
            return None

    @property
    def in_main_branch(self) -> bool:
        """
        Return True if the commit is in the main branch, False otherwise.

        :return: bool in_main_branch
        """
        return self._conf.get('main_branch') in self.branches

    @property
    def branches(self) -> Set[str]:
        """
        Return the set of branches that contain the commit.

        :return: set(str) branches
        """
        if self._branches is None:
            self._branches = self._get_branches()

        return self._branches

    def _get_branches(self):
        c_git = Git(str(self._conf.get('path_to_repo')))
        branches = set()
        for branch in set(c_git.branch('--contains', self.hash).split('\n')):
            branches.add(branch.strip().replace('* ', ''))
        return branches

    # pylint disable=R0902
    @staticmethod
    def _from_change_to_modification_type(diff: Diff):
        if diff.new_file:
            return ModificationType.ADD
        if diff.deleted_file:
            return ModificationType.DELETE
        if diff.renamed_file:
            return ModificationType.RENAME
        if diff.a_blob and diff.b_blob and diff.a_blob != diff.b_blob:
            return ModificationType.MODIFY

        return ModificationType.UNKNOWN

    def __eq__(self, other):
        if not isinstance(other, Commit):
            return NotImplemented
        if self is other:
            return True

        return self.__dict__ == other.__dict__



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module includes 1 class, RepositoryMining, main class of PyDriller.
"""

import logging
import os
import tempfile
from datetime import datetime
from typing import List, Generator, Union

from git import Repo

from pydriller.domain.commit import Commit
from pydriller.git_repository import GitRepository
from pydriller.utils.conf import Conf

logger = logging.getLogger(__name__)


class RepositoryMining:
    """
    This is the main class of PyDriller, responsible for running the study.
    """

    # pylint: disable=R0902,R0913,R0914
    def __init__(self, path_to_repo: Union[str, List[str]],
                 single: str = None,
                 since: datetime = None, to: datetime = None,
                 from_commit: str = None, to_commit: str = None,
                 from_tag: str = None, to_tag: str = None,
                 reversed_order: bool = False,
                 only_in_branch: str = None,
                 only_modifications_with_file_types: List[str] = None,
                 only_no_merge: bool = False,
                 only_authors: List[str] = None,
                 only_commits: List[str] = None,
                 only_releases: bool = False,
                 filepath: str = None,
                 histogram_diff: bool = False,
                 skip_whitespaces: bool = False):
        """
        Init a repository mining. The only required parameter is
        "path_to_repo": to analyze a single repo, pass the absolute path to
        the repo; if you need to analyze more repos, pass a list of absolute
        paths.

        Furthermore, PyDriller supports local and remote repositories: if
        you pass a path to a repo, PyDriller will run the study on that
        repo; if you pass an URL, PyDriller will clone the repo in a
        temporary folder, run the study, and delete the temporary folder.

        :param Union[str,List[str]] path_to_repo: absolute path (or list of
            absolute paths) to the repository(ies) to analyze
        :param str single: hash of a single commit to analyze
        :param datetime since: starting date
        :param datetime to: ending date
        :param str from_commit: starting commit (only if `since` is None)
        :param str to_commit: ending commit (only if `to` is None)
        :param str from_tag: starting the analysis from specified tag (only
            if `since` and `from_commit` are None)
        :param str to_tag: ending the analysis from specified tag (only if
            `to` and `to_commit` are None)
        :param bool reversed_order: whether the commits should be analyzed
            in reversed order
        :param str only_in_branch: only commits in this branch will be analyzed
        :param List[str] only_modifications_with_file_types: only
            modifications with that file types will be analyzed
        :param bool only_no_merge: if True, merges will not be analyzed
        :param List[str] only_authors: only commits of these authors will be
            analyzed (the check is done on the username, NOT the email)
        :param List[str] only_commits: only these commits will be analyzed
        :param str filepath: only commits that modified this file will be
            analyzed
        """
        options = {
            "git_repo": None,
            "path_to_repo": path_to_repo,
            "from_commit": from_commit,
            "to_commit": to_commit,
            "from_tag": from_tag,
            "to_tag": to_tag,
            "since": since,
            "to": to,
            "single": single,
            "reversed_order": reversed_order,
            "only_in_branch": only_in_branch,
            "only_modifications_with_file_types":
                only_modifications_with_file_types,
            "only_no_merge": only_no_merge,
            "only_authors": only_authors,
            "only_commits": only_commits,
            "only_releases": only_releases,
            "skip_whitespaces": skip_whitespaces,
            "filepath": filepath,
            "filepath_commits": None,
            "tagged_commits": None,
            "histogram": histogram_diff
        }
        self._conf = Conf(options)

    @staticmethod
    def _is_remote(repo: str) -> bool:
        return repo.startswith("git@") or repo.startswith("https://")

    def _clone_remote_repos(self, tmp_folder: str, repo: str) -> str:

        repo_folder = os.path.join(tmp_folder,
                                   self._get_repo_name_from_url(repo))
        logger.info("Cloning %s in temporary folder %s", repo, repo_folder)
        Repo.clone_from(url=repo, to_path=repo_folder)

        return repo_folder

    def traverse_commits(self) -> Generator[Commit, None, None]:
        """
        Analyze all the specified commits (all of them by default), returning
        a generator of commits.
        """
        for path_repo in self._conf.get('path_to_repos'):
            # if it is a remote repo, clone it first in a temporary folder!
            if self._is_remote(path_repo):
                tmp_folder = tempfile.TemporaryDirectory()
                path_repo = self._clone_remote_repos(tmp_folder.name,
                                                     path_repo)

            git_repo = GitRepository(path_repo, self._conf)
            self._conf.put("git_repo", git_repo)
            self._conf.sanity_check_filters()

            logger.info('Analyzing git repository in %s', git_repo.path)

            if self._conf.get('filepath') is not None:
                self._conf.put('filepath_commits',
                               git_repo.get_commits_modified_file(
                                   self._conf.get('filepath')))

            if self._conf.get('only_releases'):
                self._conf.put('tagged_commits', git_repo.get_tagged_commits())

            for commit in git_repo.get_list_commits(self._conf.get(
                    'only_in_branch'), not self._conf.get('reversed_order')):
                logger.info('Commit #%s in %s from %s', commit.hash,
                            commit.committer_date,
                            commit.author.name)

                if self._conf.is_commit_filtered(commit):
                    logger.info('Commit #%s filtered', commit.hash)
                    continue

                yield commit

    @staticmethod
    def _get_repo_name_from_url(url: str) -> str:
        last_slash_index = url.rfind("/")
        last_suffix_index = url.rfind(".git")
        if last_suffix_index < 0:
            last_suffix_index = len(url)

        if last_slash_index < 0 or last_suffix_index <= last_slash_index:
            raise Exception("Badly formatted url {}".format(url))

        return url[last_slash_index + 1:last_suffix_index]



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)

import pytest
from pydriller.repository_mining import RepositoryMining
from datetime import datetime, timezone, timedelta

to_zone = timezone(timedelta(hours=1))
dt = datetime(2018, 3, 22, 10, 41, 30, tzinfo=to_zone)
dt1 = datetime(2018, 3, 22, 10, 42, 3, tzinfo=to_zone)
dt2 = datetime(2018, 3, 22, 10, 41, 45, tzinfo=to_zone)
to_zone = timezone(timedelta(hours=2))
dt3 = datetime(2018, 3, 27, 17, 20, 3, tzinfo=to_zone)


@pytest.fixture
def path():
    return 'test-repos/test1/'


@pytest.fixture
def since():
    return None


@pytest.fixture
def to():
    return None


@pytest.fixture
def from_commit():
    return None


@pytest.fixture
def to_commit():
    return None


@pytest.fixture
def from_tag():
    return None


@pytest.fixture
def to_tag():
    return None


@pytest.fixture
def repository_mining_st(path, since, to):
    return list(RepositoryMining(path, since=since, to=to).traverse_commits())


@pytest.fixture
def repository_mining_cc(path, from_commit, to_commit):
    return list(RepositoryMining(path, from_commit=from_commit, to_commit=to_commit).traverse_commits())


@pytest.fixture
def repository_mining_tt(path, from_tag, to_tag):
    return list(RepositoryMining(path, from_tag=from_tag, to_tag=to_tag).traverse_commits())


@pytest.mark.parametrize('to,expected_commits', [
    (None, 5),
    (dt1, 3),
])
def test_to_filter(repository_mining_st, expected_commits):
    assert len(repository_mining_st) == expected_commits


@pytest.mark.parametrize('since,expected_commits', [
    (None, 5),
    (dt, 4),
])
def test_since_filter(repository_mining_st, expected_commits):
    assert len(repository_mining_st) == expected_commits


@pytest.mark.parametrize('since,to,expected_commits', [
    (dt2, dt3, 3),
])
def test_since_and_to_filters(repository_mining_st, expected_commits):
    assert len(repository_mining_st) == expected_commits


# FROM AND TO COMMIT
@pytest.mark.parametrize('to_commit,expected_commits', [
    ('09f6182cef737db02a085e1d018963c7a29bde5a', 3),
])
def test_to_commit_filter_new(repository_mining_cc, expected_commits):
    assert len(repository_mining_cc) == expected_commits


@pytest.mark.parametrize('from_commit,expected_commits', [
    ('6411e3096dd2070438a17b225f44475136e54e3a', 4),
])
def test_from_commit_filter(repository_mining_cc, expected_commits):
    assert len(repository_mining_cc) == expected_commits


@pytest.mark.parametrize('from_commit,to_commit,expected_commits', [
    ('6411e3096dd2070438a17b225f44475136e54e3a', '09f6182cef737db02a085e1d018963c7a29bde5a', 2),
])
def test_from_and_to_commit(repository_mining_cc, expected_commits):
    assert len(repository_mining_cc) == expected_commits


# FROM AND TO TAG
@pytest.mark.parametrize('to_tag,expected_commits', [
    ('v1.4', 3),
])
def test_to_tag_filter_new(repository_mining_tt, expected_commits):
    assert len(repository_mining_tt) == expected_commits


@pytest.mark.parametrize('from_tag,expected_commits', [
    ('v1.4', 3),
])
def test_from_tag_filter(repository_mining_tt, expected_commits):
    assert len(repository_mining_tt) == expected_commits


@pytest.mark.parametrize('from_tag,to_tag,expected_commits', [
    ('v1.4', 'v1.4', 1),
])
def test_from_and_to_tag(repository_mining_tt, expected_commits):
    assert len(repository_mining_tt) == expected_commits


def test_multiple_filters_exceptions():
    from_commit = '6411e3096dd2070438a17b225f44475136e54e3a'
    from_tag = 'v1.4'

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       since=dt2,
                                       from_commit=from_commit
                                       ).traverse_commits():
            print(commit.hash)

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       since=dt2,
                                       from_tag=from_tag).traverse_commits():
            print(commit.hash)

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       from_commit=from_commit,
                                       from_tag=from_tag).traverse_commits():
            print(commit.hash)

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       to=dt2,
                                       to_commit=from_commit
                                       ).traverse_commits():
            print(commit.hash)

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       to=dt2,
                                       to_tag=from_tag).traverse_commits():
            print(commit.hash)

    with pytest.raises(Exception):
        for commit in RepositoryMining('test-repos/test1/',
                                       single=from_commit,
                                       to=dt2,
                                       to_tag=from_tag).traverse_commits():
            print(commit.hash)



from pydriller.metrics.process.process_metrics import ProcessMetrics

def test_distinct_devs_from_hash():
    count = ProcessMetrics().distinct_dev_count(
        path_to_repo='test-repos/git-1', filepath='unexisting.java',
        to_commit='ffccf1e7497eb8136fd66ed5e42bef29677c4b71')
    assert count == 0

    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Matricula.java', to_commit='ffccf1e7497eb8136fd66ed5e42bef29677c4b71')
    assert count == 1

def test_distinct_devs_from_latest():
    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='unexisting.java')
    assert count == 0
    
    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Arquivo.java')
    assert count == 1

    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Matricula.java')
    assert count == 1

def test_distinct_devs_with_renaming():
    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Matricula.javax')
    assert count == 1

    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Secao.javax')
    assert count == 1

def test_distinct_devs_til_creation_file():
    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Secao.java')
    assert count == 1

    count = ProcessMetrics().distinct_dev_count(path_to_repo='test-repos/git-1', filepath='Secao.java', to_commit='f0dd1308bd904a9b108a6a40865166ee962af3d4')
    assert count == 1

def test_distinct_devs_from_latest_remote():
    count = ProcessMetrics().distinct_dev_count(path_to_repo='https://github.com/ishepard/pydriller', filepath='domain/developer.py', to_commit='fdf671856b260aca058e6595a96a7a0fba05454b')
    assert count == 1

def test_distinct_devs_from_latest_remote2():
    count = ProcessMetrics().distinct_dev_count(path_to_repo='https://github.com/ishepard/pydriller', filepath='domain/developer.py')
    assert count == 1



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pytest
from datetime import datetime, timezone, timedelta

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s',
                    level=logging.INFO)

from pydriller.repository_mining import RepositoryMining


def test_mod_with_file_types():
    lc = list(RepositoryMining('test-repos/git-4/',
                               only_modifications_with_file_types=['.java'])
              .traverse_commits())

    assert len(lc) == 2
    assert lc[0].hash == 'a1b6136f978644ff1d89816bc0f2bd86f6d9d7f5'
    assert lc[1].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'

    lc = list(RepositoryMining('test-repos/git-7/',
                               only_modifications_with_file_types=['.java'])
              .traverse_commits())

    assert len(lc) == 2
    assert lc[0].hash == '5adbb71167e79ab6b974827e74c9da4d81977655'
    assert lc[1].hash == '0577bec2387ee131e1ccf336adcc172224d3f6f9'


def test_mod_with_file_types_no_extension():
    lc = list(RepositoryMining('test-repos/git-4/',
                               only_modifications_with_file_types=['.py'])
              .traverse_commits())

    assert len(lc) == 0


def test_mod_with_file_types_and_date():
    to_zone = timezone(timedelta(hours=2))
    dt1 = datetime(2016, 10, 8, 23, 57, 49, tzinfo=to_zone)
    print(dt1)
    lc = list(RepositoryMining('test-repos/git-4/',
                               only_modifications_with_file_types=['.java'],
                               since=dt1)
              .traverse_commits())

    print(lc)
    assert len(lc) == 1
    assert lc[0].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'


def test_only_in_main_branch():
    lc = list(RepositoryMining('test-repos/git-5/').traverse_commits())

    assert len(lc) == 5
    assert lc[0].hash == '4a17f31c0d1285477a3a467d0bc3cb38e775097d'
    assert lc[1].hash == 'ff663cf1931a67d5e47b75fc77dcea432c728052'
    assert lc[2].hash == 'fa8217c324e7fb46c80e1ddf907f4e141449637e'
    assert lc[3].hash == '5d9d79607d7e82b6f236aa29be4ba89a28fb4f15'
    assert lc[4].hash == '377e0f474d70f6205784d0150ee0069a050c29ed'


def test_only_no_merge():
    lc = list(RepositoryMining('test-repos/git-5/',
                               only_no_merge=True).traverse_commits())

    assert len(lc) == 4
    assert lc[0].hash == '4a17f31c0d1285477a3a467d0bc3cb38e775097d'
    assert lc[1].hash == 'ff663cf1931a67d5e47b75fc77dcea432c728052'
    assert lc[2].hash == 'fa8217c324e7fb46c80e1ddf907f4e141449637e'
    assert lc[3].hash == '377e0f474d70f6205784d0150ee0069a050c29ed'


def test_no_filters():
    lc = list(RepositoryMining('test-repos/git-4/').traverse_commits())

    assert len(lc) == 3
    assert lc[0].hash == 'a1b6136f978644ff1d89816bc0f2bd86f6d9d7f5'
    assert lc[1].hash == '375de7a8275ecdc0b28dc8de2568f47241f443e9'
    assert lc[2].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'


def test_only_in_branch():
    lc = list(RepositoryMining('test-repos/git-5/',
                               only_in_branch='branch2').traverse_commits())
    assert len(lc) == 6

    assert lc[0].hash == '4a17f31c0d1285477a3a467d0bc3cb38e775097d'
    assert lc[1].hash == 'ff663cf1931a67d5e47b75fc77dcea432c728052'
    assert lc[2].hash == 'fa8217c324e7fb46c80e1ddf907f4e141449637e'
    assert lc[3].hash == '5d9d79607d7e82b6f236aa29be4ba89a28fb4f15'
    assert lc[4].hash == '377e0f474d70f6205784d0150ee0069a050c29ed'
    assert lc[5].hash == '6fe83d9fbf9a63cc1c51e5fe6fd5230f7fbbce6f'


def test_only_in_branches():
    # by default, only analyze master
    assert len(list(RepositoryMining('test-repos/test8/')
                    .traverse_commits())) == 3
    # only analyze b2
    assert len(list(RepositoryMining('test-repos/test8/', only_in_branch='b2')
                    .traverse_commits())) == 4
    # only analyze b1
    assert len(list(RepositoryMining('test-repos/test8/', only_in_branch='b1')
                    .traverse_commits())) == 5


def test_only_in_branch_not_exist():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/git-5/', only_in_branch='branch7')
             .traverse_commits())


def test_only_authors():
    lc = list(RepositoryMining('test-repos/git-10/',
                               only_authors=["Maurício Aniche"])
              .traverse_commits())
    assert len(lc) == 4

    lc = list(RepositoryMining('test-repos/git-10/',
                               only_authors=["ishepard"])
              .traverse_commits())
    assert len(lc) == 1


def test_only_authors_not_existing():
    lc = list(RepositoryMining('test-repos/git-10/',
                               only_authors=["Uncle Bob"])
              .traverse_commits())
    assert len(lc) == 0


def test_only_commits():
    # 4e669cb4f69245dc669e116517d80d038d8e0434
    # 29e929fbc5dc6a2e9c620069b24e2a143af4285f
    # 8986af2a679759e5a15794f6d56e6d46c3f302f1
    # 8169f76a3d7add54b4fc7bca7160d1f1eede6eda
    # 168b3aab057ed61a769acf336a4ef5e64f76c9fd
    lc = list(RepositoryMining('test-repos/git-10/',
                               only_commits=["4e669cb4f69245dc669e116517d80d038d8e0434"]).traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "4e669cb4f69245dc669e116517d80d038d8e0434"

    lc = list(RepositoryMining('test-repos/git-10/',
                               only_commits=["4e669cb4f69245dc669e116517d80d038d8e0434",
                                             "8986af2a679759e5a15794f6d56e6d46c3f302f1"]).traverse_commits())
    assert len(lc) == 2
    assert lc[0].hash == "8986af2a679759e5a15794f6d56e6d46c3f302f1"
    assert lc[1].hash == "4e669cb4f69245dc669e116517d80d038d8e0434"

    lc = list(RepositoryMining('test-repos/git-10/',
                               only_commits=["4e669cb4f69245dc669e116517d80d038d8e0434",
                                             "8986af2a679759e5a15794f6d56e6d46c3f302f1",
                                             "29e929fbc5dc6a2e9c620069b24e2a143af4285f"]).traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == "8986af2a679759e5a15794f6d56e6d46c3f302f1"
    assert lc[1].hash == "29e929fbc5dc6a2e9c620069b24e2a143af4285f"
    assert lc[2].hash == "4e669cb4f69245dc669e116517d80d038d8e0434"

    lc = list(RepositoryMining('test-repos/git-10/',
                               only_commits=["fake hash"]).traverse_commits())
    assert len(lc) == 0

    only_commits = len(list(RepositoryMining('test-repos/git-10/',
                                             only_commits=["4e669cb4f69245dc669e116517d80d038d8e0434",
                                                           "29e929fbc5dc6a2e9c620069b24e2a143af4285f",
                                                           "8986af2a679759e5a15794f6d56e6d46c3f302f1",
                                                           "8169f76a3d7add54b4fc7bca7160d1f1eede6eda",
                                                           "168b3aab057ed61a769acf336a4ef5e64f76c9fd"]).traverse_commits()))

    total_commits = len(list(RepositoryMining('test-repos/git-10/').traverse_commits()))

    assert total_commits == only_commits


def test_single_commit():
    lc = list(RepositoryMining('test-repos/git-10/',
                               single="4e669cb4f69245dc669e116517d80d038d8e0434").traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "4e669cb4f69245dc669e116517d80d038d8e0434"

    lc = list(RepositoryMining('test-repos/git-10/',
                               single="168b3aab057ed61a769acf336a4ef5e64f76c9fd").traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "168b3aab057ed61a769acf336a4ef5e64f76c9fd"


def test_filepath_with_to():
    dt = datetime(2018, 6, 6)
    assert len(list(RepositoryMining(
        path_to_repo='test-repos/test5',
        filepath='myfolder/A.java',
        to=dt).traverse_commits())) == 4


def test_filepath_with_since():
    since = datetime(2018, 6, 6)
    assert len(list(RepositoryMining(
        path_to_repo='test-repos/test5',
        filepath='myfolder/A.java',
        since=since).traverse_commits())) == 10


def test_filepath_with_rename():
    dt = datetime(2018, 6, 6)
    commits = list(RepositoryMining(
        path_to_repo='test-repos/test1',
        filepath='file4.java',
        to=dt).traverse_commits())
    assert len(commits) == 2

    commit_hashes = [commit.hash for commit in commits]

    assert 'da39b1326dbc2edfe518b90672734a08f3c13458' in commit_hashes
    assert 'a88c84ddf42066611e76e6cb690144e5357d132c' in commit_hashes


def test_filepath_with_rename_complex():
    commits = list(RepositoryMining(
        path_to_repo='test-repos/git-1',
        filepath='Matricula.javax').traverse_commits())
    assert len(commits) == 6

    commit_hashes = [commit.hash for commit in commits]

    assert 'f0dd1308bd904a9b108a6a40865166ee962af3d4' in commit_hashes
    assert '953737b199de233896f00b4d87a0bc2794317253' in commit_hashes
    assert 'a3290ac2f555eabca9e31180cf38e91f9e7e2761' in commit_hashes
    assert '71535a31f0b598a5d5fcebda7146ebc01def783a' in commit_hashes
    assert '57dbd017d1a744b949e7ca0b1c1a3b3dd4c1cbc1' in commit_hashes
    assert '866e997a9e44cb4ddd9e00efe49361420aff2559' in commit_hashes


def test_only_releases():
    lc = list(RepositoryMining('test-repos/git-8/',
                               only_releases=True).traverse_commits())

    assert len(lc) == 3
    assert '6bb9e2c6a8080e6b5b34e6e316c894b2ddbf7fcd' == lc[0].hash
    assert '4638730126d40716e230c2040751a13153fb1556' == lc[1].hash
    assert '627e1ad917a188a861c9fedf6e5858b79edbe439' == lc[2].hash


def test_only_releases_wo_releases():
    lc = list(RepositoryMining('test-repos/git-1/',
                               only_releases=True).traverse_commits())

    assert len(lc) == 0


"""
Configuration module.
"""

import logging
from datetime import datetime

import pytz

from pydriller.domain.commit import Commit

logger = logging.getLogger(__name__)


class Conf:
    """
    Configuration class. This class holds all the possible configurations of
    the mining process (i.e., starting and ending dates, branches, etc.)
    It's also responsible for checking whether the filters are correct (i.e.,
    the user did not specify 2 starting commits).
    """
    def __init__(self, options):
        # insert all the configurations in a local dictionary
        self._options = {}
        for key, val in options.items():
            self._options[key] = val

        self._sanity_check_repos(self.get('path_to_repo'))
        if isinstance(self.get('path_to_repo'), str):
            self.put('path_to_repos', [self.get('path_to_repo')])
        else:
            self.put('path_to_repos', self.get('path_to_repo'))

    def put(self, key, value):
        """
        Save the value of a configuration.

        :param key: configuration (i.e., start date)
        :param value: value
        """
        self._options[key] = value

    def get(self, key):
        """
        Return the value of the configuration.

        :param key: configuration name
        :return: value of the configuration, None if not present
        """
        return self._options.get(key, None)

    @staticmethod
    def _sanity_check_repos(path_to_repo):
        """
        Checks if repo is of type str or list.

        @param path_to_repo: path to the repo as provided by the user.
        @return:
        """
        if not isinstance(path_to_repo, str) and \
                not isinstance(path_to_repo, list):
            raise Exception("The path to the repo has to be of type "
                            "'string' or 'list of strings'!")

    def sanity_check_filters(self):
        """
        Check if the values passed by the user are correct.

        """
        if self.get('single') is not None:
            if any([self.get('since'),
                    self.get('to'),
                    self.get('from_commit'),
                    self.get('to_commit'),
                    self.get('from_tag'),
                    self.get('to_tag')]):
                raise Exception('You can not specify a single commit with '
                                'other filters')

        self.check_starting_commit()
        self.check_ending_commit()
        self._check_timezones()

    def check_ending_commit(self):
        """
        Get the ending commit from the 'to', 'to_commit' or 'to_tag' filter.
        """
        if not self.only_one_filter([self.get('to'),
                                     self.get('to_commit'),
                                     self.get('to_tag')]):
            raise Exception('You can only specify one between since, '
                            'from_tag and from_commit')
        git_repo = self.get("git_repo")
        if self.get('to_commit') is not None:
            self.put('to', git_repo.get_commit(self.get(
                'to_commit')).committer_date)
        if self.get('to_tag') is not None:
            self.put('to', git_repo.get_commit_from_tag(
                self.get('to_tag')).committer_date)

    def check_starting_commit(self):
        """
        Get the starting commit from the 'since', 'from_commit' or 'from_tag'
        filter.
        """
        if not self.only_one_filter([self.get('since'),
                                     self.get('from_commit'),
                                     self.get('from_tag')]):
            raise Exception('You can only specify one between since, '
                            'from_tag and from_commit')
        git_repo = self.get("git_repo")
        if self.get('from_commit') is not None:
            self.put('since', git_repo.get_commit(
                self.get('from_commit')).committer_date)
        if self.get('from_tag') is not None:
            self.put('since', git_repo.get_commit_from_tag(
                self.get('from_tag')).committer_date)

    @staticmethod
    def only_one_filter(arr):
        """
        Return true if in 'arr' there is at most 1 filter to True.

        :param arr: iterable object
        :return:
        """
        return len([x for x in arr if x is not None]) <= 1

    def is_commit_filtered(self, commit: Commit):  # pylint: disable=R0911
        """
        Check if commit has to be filtered according to the filters provided
        by the user.

        :param Commit commit: Commit to check
        :return:
        """
        if self.get('single') is not None and \
                commit.hash != self.get('single'):
            logger.debug(
                'Commit filtered because is not the defined in single')
            return True
        if (self.get('since') is not None and
                commit.committer_date < self.get('since')) or \
                (self.get('to') is not None and
                 commit.committer_date > self.get('to')):
            return True
        if self.get('only_modifications_with_file_types') is not None:
            if not self._has_modification_with_file_type(commit):
                logger.debug('Commit filtered for modification types')
                return True
        if self.get('only_no_merge') is True and commit.merge is True:
            logger.debug('Commit filtered for no merge')
            return True
        if self.get('only_authors') is not None and \
                commit.author.name not in self.get('only_authors'):
            logger.debug("Commit filtered for author")
            return True
        if self.get('only_commits') is not None and \
                commit.hash not in self.get('only_commits'):
            logger.debug("Commit filtered because it is not one of the "
                         "specified commits")
            return True
        if self.get('filepath_commits') is not None and \
                commit.hash not in self.get('filepath_commits'):
            logger.debug("Commit filtered because it did not modify the "
                         "specified file")
            return True
        if self.get('tagged_commits') is not None and \
                commit.hash not in self.get('tagged_commits'):
            logger.debug("Commit filtered because it is not tagged")
            return True

        return False

    def _has_modification_with_file_type(self, commit):
        for mod in commit.modifications:
            if mod.filename.endswith(
                    tuple(self.get('only_modifications_with_file_types'))):
                return True
        return False

    def _check_timezones(self):
        if self.get('since') is not None:
            self.put('since', self._replace_timezone(self.get('since')))
        if self.get('to') is not None:
            self.put('to', self._replace_timezone(self.get('to')))

    @staticmethod
    def _replace_timezone(dt: datetime):
        if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
            dt = dt.replace(tzinfo=pytz.utc)
        return dt



import logging
from datetime import datetime

import pytest

from pydriller import RepositoryMining, GitRepository

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s',
                    level=logging.INFO)


# It should fail when no URLs are specified
def test_no_url():
    with pytest.raises(Exception):
        list(RepositoryMining().traverse_commits())


# It should fail when URL is not a string or a List
def test_badly_formatted_repo_url():
    with pytest.raises(Exception):
        list(RepositoryMining(path_to_repo=set('repo')).traverse_commits())


def test_simple_url():
    assert len(list(RepositoryMining(
        path_to_repo="test-repos/test1").traverse_commits())) == 5


def test_two_local_urls():
    urls = ["test-repos/test1", "test-repos/test3"]
    assert len(list(RepositoryMining(
        path_to_repo=urls).traverse_commits())) == 11


def test_simple_remote_url():
    dt2 = datetime(2018, 10, 20)
    assert len(list(RepositoryMining(
        path_to_repo="https://github.com/ishepard/pydriller.git",
        to=dt2).traverse_commits())) == 159


def test_two_remote_urls():
    urls = ["https://github.com/mauricioaniche/repodriller.git",
            "https://github.com/ishepard/pydriller"]
    dt2 = datetime(2018, 10, 20)
    assert len(list(RepositoryMining(path_to_repo=urls,
                                     to=dt2).traverse_commits())) == 518


def test_2_identical_local_urls():
    urls = ["test-repos/test1", "test-repos/test1"]
    assert len(list(RepositoryMining(
        path_to_repo=urls).traverse_commits())) == 10


def test_both_local_and_remote_urls():
    dt2 = datetime(2018, 10, 20)
    assert len(list(RepositoryMining(
        path_to_repo=["test-repos/test1",
                      "https://github.com/ishepard/pydriller.git"],
        to=dt2).traverse_commits())) == 164


def test_both_local_and_remote_urls_list():
    dt2 = datetime(2018, 10, 20)
    urls = ["test-repos/test1",
            "https://github.com/mauricioaniche/repodriller.git",
            "test-repos/test3",
            "https://github.com/ishepard/pydriller.git"]
    assert len(list(RepositoryMining(path_to_repo=urls,
                                     to=dt2).traverse_commits())) == 529


def test_badly_formatted_url():
    with pytest.raises(Exception):
        list(RepositoryMining(
            path_to_repo='https://github.com/ishepard.git/test')
             .traverse_commits())

    with pytest.raises(Exception):
        list(RepositoryMining(path_to_repo='test').traverse_commits())


def test_diff_histogram():
    # without histogram
    commit = list(RepositoryMining('test-repos/test13',
                                   single="93df8676e6fab70d9677e94fd0f6b17db095e890").traverse_commits())[0]
    mod = commit.modifications[0]
    gr = GitRepository('test-repos/test13')
    diff = gr.parse_diff(mod.diff)
    assert len(diff['added']) == 11
    assert (3, '    if (path == null)') in diff['added']
    assert (5, '        log.error("Icon path is null");') in diff['added']
    assert (6, '        return null;') in diff['added']
    assert (8, '') in diff['added']
    assert (9, '    java.net.URL imgURL = GuiImporter.class.getResource(path);') in diff['added']
    assert (10, '') in diff['added']
    assert (11, '    if (imgURL == null)') in diff['added']
    assert (12, '    {') in diff['added']
    assert (14, '        return null;') in diff['added']
    assert (16, '    else') in diff['added']
    assert (17, '        return new ImageIcon(imgURL);') in diff['added']

    assert len(diff['deleted']) == 7
    assert (3, '    java.net.URL imgURL = GuiImporter.class.getResource(path);') in diff['deleted']
    assert (4, '') in diff['deleted']
    assert (5, '    if (imgURL != null)') in diff['deleted']
    assert (7, '        return new ImageIcon(imgURL);') in diff['deleted']
    assert (9, '    else') in diff['deleted']
    assert (10, '    {') in diff['deleted']
    assert (13, '    return null;') in diff['deleted']

    # with histogram
    commit = list(RepositoryMining('test-repos/test13',
                                   single="93df8676e6fab70d9677e94fd0f6b17db095e890",
                                   histogram_diff=True).traverse_commits())[0]
    mod = commit.modifications[0]
    gr = GitRepository('test-repos/test13')
    diff = gr.parse_diff(mod.diff)
    assert (4, '    {') in diff["added"]
    assert (5, '        log.error("Icon path is null");') in diff["added"]
    assert (6, '        return null;') in diff["added"]
    assert (7, '    }') in diff["added"]
    assert (8, '') in diff["added"]
    assert (11, '    if (imgURL == null)') in diff["added"]
    assert (12, '    {') in diff["added"]
    assert (13, '        log.error("Couldn\'t find icon: " + imgURL);') in diff["added"]
    assert (14, '        return null;') in diff["added"]
    assert (17, '        return new ImageIcon(imgURL);') in diff["added"]

    assert (6, '    {') in diff["deleted"]
    assert (7, '        return new ImageIcon(imgURL);') in diff["deleted"]
    assert (10, '    {') in diff["deleted"]
    assert (11, '        log.error("Couldn\'t find icon: " + imgURL);') in diff["deleted"]
    assert (12, '    }') in diff["deleted"]
    assert (13, '    return null;') in diff["deleted"]


def test_ignore_add_whitespaces():
    commit = list(RepositoryMining('test-repos/test14',
                                   single="338a74ceae164784e216555d930210371279ba8e").traverse_commits())[0]
    assert len(commit.modifications) == 1
    commit = list(RepositoryMining('test-repos/test14',
                                   skip_whitespaces=True,
                                   single="338a74ceae164784e216555d930210371279ba8e").traverse_commits())[0]
    assert len(commit.modifications) == 0


def test_ignore_add_whitespaces_and_modified_normal_line():
    gr = GitRepository('test-repos/test14')
    commit = list(RepositoryMining('test-repos/test14',
                                   single="52716ef1f11e07308b5df1b313aec5496d5e91ce").traverse_commits())[0]
    assert len(commit.modifications) == 1
    parsed_normal_diff = gr.parse_diff(commit.modifications[0].diff)
    commit = list(RepositoryMining('test-repos/test14',
                                   skip_whitespaces=True,
                                   single="52716ef1f11e07308b5df1b313aec5496d5e91ce").traverse_commits())[0]
    assert len(commit.modifications) == 1
    parsed_wo_whitespaces_diff = gr.parse_diff(commit.modifications[0].diff)
    assert len(parsed_normal_diff['added']) == 2
    assert len(parsed_wo_whitespaces_diff['added']) == 1

    assert len(parsed_normal_diff['deleted']) == 1
    assert len(parsed_wo_whitespaces_diff['deleted']) == 0


def test_ignore_deleted_whitespaces():
    commit = list(RepositoryMining('test-repos/test14',
                                   single="e6e429f6b485e18fb856019d9953370fd5420b20").traverse_commits())[0]
    assert len(commit.modifications) == 1
    commit = list(RepositoryMining('test-repos/test14',
                                   skip_whitespaces=True,
                                   single="e6e429f6b485e18fb856019d9953370fd5420b20").traverse_commits())[0]
    assert len(commit.modifications) == 0


def test_ignore_add_whitespaces_and_changed_file():
    commit = list(RepositoryMining('test-repos/test14',
                                   single="532068e9d64b8a86e07eea93de3a57bf9e5b4ae0").traverse_commits())[0]
    assert len(commit.modifications) == 2
    commit = list(RepositoryMining('test-repos/test14',
                                   skip_whitespaces=True,
                                   single="532068e9d64b8a86e07eea93de3a57bf9e5b4ae0").traverse_commits())[0]
    assert len(commit.modifications) == 1



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module includes 1 class, GitRepository, representing a repository in Git.
"""

import logging
import os
from pathlib import Path
from threading import Lock
from typing import List, Dict, Tuple, Set, Generator

from git import Git, Repo, GitCommandError, Commit as GitCommit

from pydriller.domain.commit import Commit, ModificationType, Modification
from pydriller.utils.conf import Conf
from pydriller.utils.hyperblame import GitHyperBlame

logger = logging.getLogger(__name__)


class GitRepository:
    """
    Class representing a repository in Git. It contains most of the logic of
    PyDriller: obtaining the list of commits, checkout, reset, etc.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, path: str, conf=None):
        """
        Init the Git RepositoryMining.

        :param str path: path to the repository
        """
        self.path = Path(path)
        self.hyperblame = GitHyperBlame(path)
        self.project_name = self.path.name
        self.lock = Lock()
        self._hyper_blame_available = None
        self._git = None
        self._repo = None

        # if no configuration is passed, then creates a new "emtpy" one
        # with just "path_to_repo" inside.
        if conf is None:
            conf = Conf({
                "path_to_repo": str(self.path),
                "git_repo": self
            })

        self._conf = conf
        self._conf.put("main_branch", None)  # init main_branch to None

    @property
    def git(self):
        """
        GitPython object Git.

        :return: Git
        """
        if self._git is None:
            self._open_git()
        return self._git

    @property
    def repo(self):
        """
        GitPython object Repo.

        :return: Repo
        """
        if self._repo is None:
            self._open_repository()
        return self._repo

    def _open_git(self):
        self._git = Git(str(self.path))

    def _open_repository(self):
        self._repo = Repo(str(self.path))
        if self._conf.get("main_branch") is None:
            self._discover_main_branch(self._repo)

    def _discover_main_branch(self, repo):
        try:
            self._conf.put("main_branch", repo.active_branch.name)
        except TypeError:
            logger.info("HEAD is a detached symbolic reference, setting "
                        "main branch to empty string")
            self._conf.put("main_branch", '')

    def get_head(self) -> Commit:
        """
        Get the head commit.

        :return: Commit of the head commit
        """
        head_commit = self.repo.head.commit
        return Commit(head_commit, self._conf)

    def get_list_commits(self, branch: str = None,
                         reverse_order: bool = True) \
            -> Generator[Commit, None, None]:
        """
        Return a generator of commits of all the commits in the repo.

        :return: Generator[Commit], the generator of all the commits in the
            repo
        """
        for commit in self.repo.iter_commits(branch, reverse=reverse_order):
            yield self.get_commit_from_gitpython(commit)

    def get_commit(self, commit_id: str) -> Commit:
        """
        Get the specified commit.

        :param str commit_id: hash of the commit to analyze
        :return: Commit
        """
        gp_commit = self.repo.commit(commit_id)
        return Commit(gp_commit, self._conf)

    def get_commit_from_gitpython(self, commit: GitCommit) -> Commit:
        """
        Build a PyDriller commit object from a GitPython commit object.
        This is internal of PyDriller, I don't think users generally will need
        it.

        :param GitCommit commit: GitPython commit
        :return: Commit commit: PyDriller commit
        """
        return Commit(commit, self._conf)

    def checkout(self, _hash: str) -> None:
        """
        Checkout the repo at the speficied commit.
        BE CAREFUL: this will change the state of the repo, hence it should
        *not* be used with more than 1 thread.

        :param _hash: commit hash to checkout
        """
        with self.lock:
            self._delete_tmp_branch()
            self.git.checkout('-f', _hash, b='_PD')

    def _delete_tmp_branch(self) -> None:
        try:
            # we are already in _PD, so checkout the master branch before
            # deleting it
            if self.repo.active_branch.name == '_PD':
                self.git.checkout('-f', self._conf.get("main_branch"))
            self.repo.delete_head('_PD', force=True)
        except GitCommandError:
            logger.debug("Branch _PD not found")

    def files(self) -> List[str]:
        """
        Obtain the list of the files (excluding .git directory).

        :return: List[str], the list of the files
        """
        _all = []
        for path, _, files in os.walk(str(self.path)):
            if '.git' in path:
                continue
            for name in files:
                _all.append(os.path.join(path, name))
        return _all

    def reset(self) -> None:
        """
        Reset the state of the repo, checking out the main branch and
        discarding
        local changes (-f option).

        """
        with self.lock:
            self.git.checkout('-f', self._conf.get("main_branch"))
            self._delete_tmp_branch()

    def total_commits(self) -> int:
        """
        Calculate total number of commits.

        :return: the total number of commits
        """
        return len(list(self.get_list_commits()))

    def get_commit_from_tag(self, tag: str) -> Commit:
        """
        Obtain the tagged commit.

        :param str tag: the tag
        :return: Commit commit: the commit the tag referred to
        """
        try:
            selected_tag = self.repo.tags[tag]
            return self.get_commit(selected_tag.commit.hexsha)
        except (IndexError, AttributeError):
            logger.debug('Tag %s not found', tag)
            raise

    def get_tagged_commits(self):
        """
        Obtain the hash of all the tagged commits.

        :return: list of tagged commits (can be empty if there are no tags)
        """
        tags = []
        for tag in self.repo.tags:
            if tag.commit:
                tags.append(tag.commit.hexsha)
        return tags

    def parse_diff(self, diff: str) -> Dict[str, List[Tuple[int, str]]]:
        """
        Given a diff, returns a dictionary with the added and deleted lines.
        The dictionary has 2 keys: "added" and "deleted", each containing the
        corresponding added or deleted lines. For both keys, the value is a
        list of Tuple (int, str), corresponding to (number of line in the file,
        actual line).


        :param str diff: diff of the commit
        :return: Dictionary
        """
        lines = diff.split('\n')
        modified_lines = {'added': [], 'deleted': []}

        count_deletions = 0
        count_additions = 0

        for line in lines:
            line = line.rstrip()
            count_deletions += 1
            count_additions += 1

            if line.startswith('@@'):
                count_deletions, count_additions = self._get_line_numbers(line)

            if line.startswith('-'):
                modified_lines['deleted'].append((count_deletions, line[1:]))
                count_additions -= 1

            if line.startswith('+'):
                modified_lines['added'].append((count_additions, line[1:]))
                count_deletions -= 1

            if line == r'\ No newline at end of file':
                count_deletions -= 1
                count_additions -= 1

        return modified_lines

    @staticmethod
    def _get_line_numbers(line):
        token = line.split(" ")
        numbers_old_file = token[1]
        numbers_new_file = token[2]
        delete_line_number = int(numbers_old_file.split(",")[0]
                                 .replace("-", "")) - 1
        additions_line_number = int(numbers_new_file.split(",")[0]) - 1
        return delete_line_number, additions_line_number

    def get_commits_last_modified_lines(self, commit: Commit,
                                        modification: Modification = None,
                                        hyper_blame: bool = False,
                                        hashes_to_ignore_path: str = None) \
            -> Dict[str, Set[str]]:
        """
        Given the Commit object, returns the set of commits that last
        "touched" the lines that are modified in the files included in the
        commit. It applies SZZ.

        IMPORTANT: for better results, we suggest to install Google
        depot_tools first (see
        https://dev.chromium.org/developers/how-tos/install-depot-tools).
        This allows PyDriller to use "git hyper-blame" instead of the normal
        blame. If depot_tools are not installed, PyDriller will automatically
        switch to the normal blame.

        The algorithm works as follow: (for every file in the commit)

        1- obtain the diff

        2- obtain the list of deleted lines

        3- blame the file and obtain the commits were those lines were added

        Can also be passed as parameter a single Modification, in this case
        only this file will be analyzed.

        :param Commit commit: the commit to analyze
        :param Modification modification: single modification to analyze
        :param bool hyper_blame: whether to use git hyper blame or the
            normal blame (by default it uses the normal blame).
        :param str hashes_to_ignore_path: path to a file containing hashes of
               commits to ignore. (only works with git hyper blame)
        :return: the set containing all the bug inducing commits
        """
        hashes_to_ignore = []
        if hashes_to_ignore_path is not None:
            assert os.path.exists(hashes_to_ignore_path), \
                "The file with the commit hashes to ignore does not exist"
            hashes_to_ignore = open(hashes_to_ignore_path).readlines()

        if modification is not None:
            modifications = [modification]
        else:
            modifications = commit.modifications

        return self._calculate_last_commits(commit, modifications,
                                            hyper_blame,
                                            hashes_to_ignore)

    def _calculate_last_commits(self, commit: Commit,
                                modifications: List[Modification],
                                hyper_blame: bool = False,
                                hashes_to_ignore: List[str] = None) \
            -> Dict[str, Set[str]]:

        commits = {}

        for mod in modifications:
            path = mod.new_path
            if mod.change_type == ModificationType.RENAME or \
                    mod.change_type == ModificationType.DELETE:
                path = mod.old_path
            deleted_lines = self.parse_diff(mod.diff)['deleted']
            try:
                blame = self._get_blame(commit.hash, path, hyper_blame,
                                        hashes_to_ignore)
                for num_line, line in deleted_lines:
                    if not self._useless_line(line.strip()):
                        buggy_commit = blame[num_line - 1].split(' ')[
                            0].replace('^', '')

                        if mod.change_type == ModificationType.RENAME:
                            path = mod.new_path

                        commits.setdefault(path, set()).add(
                            self.get_commit(buggy_commit).hash)
            except GitCommandError:
                logger.debug(
                    "Could not found file %s in commit %s. Probably a double "
                    "rename!", mod.filename, commit.hash)

        return commits

    def _get_blame(self, commit_hash: str, path: str,
                   hyper_blame: bool = False,
                   hashes_to_ignore: List[str] = None):
        """
        If "git hyper-blame" is available, use it. Otherwise use normal blame.
        """
        if not hyper_blame or hashes_to_ignore is None:
            return self.git.blame('-w', commit_hash + '^',
                                  '--', path).split('\n')
        return self.hyperblame.hyper_blame(hashes_to_ignore, path,
                                           commit_hash + '^')

    @staticmethod
    def _useless_line(line: str):
        # this covers comments in Java and Python, as well as empty lines.
        # More have to be added!
        return not line or \
               line.startswith('//') or \
               line.startswith('#') or \
               line.startswith("/*") or \
               line.startswith("'''") or \
               line.startswith('"""') or \
               line.startswith("*")

    def get_commits_modified_file(self, filepath: str) -> List[str]:
        """
        Given a filepath, returns all the commits that modified this file
        (following renames).

        :param str filepath: path to the file
        :return: the list of commits' hash
        """
        path = str(Path(filepath))

        commits = []
        try:
            commits = self.git.log("--follow", "--format=%H", path).split('\n')
        except GitCommandError:
            logger.debug("Could not find information of file %s", path)

        return commits



"""
Module that calculates process metrics.
See https://ieeexplore.ieee.org/document/6606589 for more info.
"""

from pathlib import Path
from pydriller.domain.commit import ModificationType
from pydriller.repository_mining import RepositoryMining


class ProcessMetrics:
    """
    This class is responsible to implement the following process metrics:

    * Commit Count: measures the number of commits made to a file
    * Distinct Developers Count: measures the cumulative number of distinct \
    developers that contributed to a file
    """

    def commits_count(self, path_to_repo: str, filepath: str,
                      from_commit: str = None, to_commit: str = None):
        """
        Return the number of commits made to a file from the first commit to
        the one identified by commit_hash.

        :path_to_repo: path to a single repo
        :commit_hash: the SHA of the commit to stop counting. If None, the
            analysis starts from the latest commit
        :filepath: the path to the file to count for. E.g. 'doc/README.md'

        :return: int number of commits made to the file
        """

        filepath = str(Path(filepath))
        count = 0

        for commit in RepositoryMining(path_to_repo, from_commit=from_commit,
                                       to_commit=to_commit,
                                       reversed_order=True).traverse_commits():
            for modified_file in commit.modifications:
                if filepath in (modified_file.new_path,
                                modified_file.old_path):
                    count += 1

                    if modified_file.change_type == ModificationType.RENAME:
                        filepath = str(Path(modified_file.old_path))

                    break
        return count

    def distinct_dev_count(self, path_to_repo: str, filepath: str,
                           from_commit: str = None, to_commit: str = None):
        """
        Return the cumulative number of distinct developers contributed to
        the file up to the indicated commit.

        :path_to_repo: path to a single repo
        :commit_hash: the SHA of the commit to stop counting. If None,
            the SHA is the latest commit SHA
        :filepath: the path to the file to count for. E.g. 'doc/README.md'

        :return: int number of distinct developers contributing to the file
        """
        filepath = str(Path(filepath))
        developers = set()

        for commit in RepositoryMining(path_to_repo, from_commit=from_commit,
                                       to_commit=to_commit,
                                       reversed_order=True).traverse_commits():

            for modified_file in commit.modifications:
                if filepath in (modified_file.new_path,
                                modified_file.old_path):
                    developers.add(commit.author.email.strip())

                    if modified_file.change_type == ModificationType.RENAME:
                        filepath = str(Path(modified_file.old_path))

                    break

        return len(developers)



"""
Copyright 2016 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.

This is a slightly modified verion of the "git hyper-blame" found in Google
depot_tools.
See https://github.com/GiantPay/depot_tools/blob/master/git_hyper_blame.py
for more information.
"""

import collections
from datetime import datetime, timedelta, timezone
import logging

from git import Git

logger = logging.getLogger(__name__)
BlameLine = collections.namedtuple(
    'BlameLine',
    'commit context lineno_then lineno_now modified')


class HyperBlameCommit():
    """Info about a commit."""
    def __init__(self, commithash):
        self.commithash = commithash
        self.author = None
        self.author_mail = None
        self.author_time = None
        self.author_tz = None
        self.committer = None
        self.committer_mail = None
        self.committer_time = None
        self.committer_tz = None
        self.summary = None
        self.boundary = None
        self.previous = None
        self.filename = None

    def __repr__(self):  # pragma: no cover
        return '<Commit %s>' % self.commithash


class GitHyperBlame:
    def __init__(self, path: str):
        self.g = Git(path)
        self.diff_hunks_cache = {}

    def parse_blame(self, blameoutput):
        """Parses the output of git blame -p into a data structure."""
        lines = blameoutput.split('\n')
        i = 0
        commits = {}

        while i < len(lines):
            # Read a commit line and parse it.
            line = lines[i]
            i += 1
            if not line.strip():
                continue
            commitline = line.split()
            commithash = commitline[0]
            lineno_then = int(commitline[1])
            lineno_now = int(commitline[2])

            try:
                commit = commits[commithash]
            except KeyError:
                commit = HyperBlameCommit(commithash)
                commits[commithash] = commit

            # Read commit details until we find a context line.
            while i < len(lines):
                line = lines[i]
                i += 1
                if line.startswith('\t'):
                    break

                try:
                    key, value = line.split(' ', 1)
                except ValueError:
                    key = line
                    value = True
                setattr(commit, key.replace('-', '_'), value)

            context = line[1:]

            yield BlameLine(commit, context, lineno_then, lineno_now, False)

    def get_parsed_blame(self, filename, revision='HEAD'):
        blame = self.g.blame('-p', revision, '--', filename)
        return list(self.parse_blame(blame))

    def hyper_blame(self, ignored, filename, revision='HEAD'):
        # Map from commit to parsed blame from that commit.
        blame_from = {}

        def cache_blame_from(filename, commithash):
            try:
                return blame_from[commithash]
            except KeyError:
                parsed = self.get_parsed_blame(filename, commithash)
                blame_from[commithash] = parsed
                return parsed

        parsed = cache_blame_from(filename, revision)

        new_parsed = []
        for line in parsed:
            # If a line references an ignored commit, blame that commit's
            # parent repeatedly until we find a non-ignored commit.
            while line.commit.commithash in ignored:
                if line.commit.previous is None:
                    # You can't ignore the commit that added this file.
                    break

                previouscommit, previousfilename = line.commit.previous.split(
                    ' ', 1)
                parent_blame = cache_blame_from(previousfilename,
                                                previouscommit)

                if len(parent_blame) == 0:
                    # The previous version of this file was empty,
                    # therefore, you can't ignore this commit.
                    break

                # line.lineno_then is the line number in question at
                # line.commit. We need
                # to translate that line number so that it refers to the
                # position of the same line on previouscommit.
                lineno_previous = self.approx_lineno_across_revs(
                    line.commit.filename, previousfilename,
                    line.commit.commithash,
                    previouscommit, line.lineno_then)
                logger.debug('ignore commit %s on line p%d/t%d/n%d',
                             line.commit.commithash, lineno_previous,
                             line.lineno_then,
                             line.lineno_now)

                # Get the line at lineno_previous in the parent commit.
                assert 1 <= lineno_previous <= len(parent_blame)
                newline = parent_blame[lineno_previous - 1]

                # Replace the commit and lineno_then, but not the lineno_now
                # or context.
                line = BlameLine(newline.commit, line.context,
                                 newline.lineno_then,
                                 line.lineno_now, True)
                logger.debug('replacing with %r', line)

            new_parsed.append(line)

        return self.build_result(new_parsed)

    def approx_lineno_across_revs(self, filename, newfilename, revision,
                                  newrevision, lineno):
        """Computes the approximate movement of a line number between two
        revisions.

        Consider line |lineno| in |filename| at |revision|. This function
        computes the
        line number of that line in |newfilename| at |newrevision|. This is
        necessarily approximate.

        Args:
          filename: The file (within the repo) at |revision|.
          newfilename: The name of the same file at |newrevision|.
          revision: A git revision.
          newrevision: Another git revision. Note: Can be ahead or behind
          |revision|.
          lineno: Line number within |filename| at |revision|.

        Returns:
          Line number within |newfilename| at |newrevision|.
        """
        # This doesn't work that well if there are a lot of line changes
        # within the
        # hunk (demonstrated by
        # GitHyperBlameLineMotionTest.testIntraHunkLineMotion).
        # A fuzzy heuristic that takes the text of the new line and tries to
        # find a
        # deleted line within the hunk that mostly matches the new line
        # could help.

        # Use the <revision>:<filename> syntax to diff between two blobs.
        # This is the only way to diff a file that has been renamed.
        old = '%s:%s' % (revision, filename)
        new = '%s:%s' % (newrevision, newfilename)
        hunks = self.cache_diff_hunks(old, new)

        cumulative_offset = 0

        # Find the hunk containing lineno (if any).
        for (oldstart, oldlength), (newstart, newlength) in hunks:
            cumulative_offset += newlength - oldlength

            if lineno >= oldstart + oldlength:
                # Not there yet.
                continue

            if lineno < oldstart:
                # Gone too far.
                break

            # lineno is in [oldstart, oldlength] at revision; [newstart,
            # newlength] at
            # newrevision.

            # If newlength == 0, newstart will be the line before the
            # deleted hunk.
            # Since the line must have been deleted, just return that as the
            # nearest
            # line in the new file. Caution: newstart can be 0 in this case.
            if newlength == 0:
                return max(1, newstart)

            newend = newstart + newlength - 1

            # Move lineno based on the amount the entire hunk shifted.
            lineno = lineno + newstart - oldstart
            # Constrain the output within the range [newstart, newend].
            return min(newend, max(newstart, lineno))

        # Wasn't in a hunk. Figure out the line motion based on the
        # difference in
        # length between the hunks seen so far.
        return lineno + cumulative_offset

    def cache_diff_hunks(self, oldrev, newrev):
        def parse_start_length(s):
            # Chop the '-' or '+'.
            s = s[1:]
            # Length is optional (defaults to 1).
            try:
                start, length = s.split(',')
            except ValueError:
                start = s
                length = 1
            return int(start), int(length)

        try:
            return self.diff_hunks_cache[(oldrev, newrev)]
        except KeyError:
            pass

        # Use -U0 to get the smallest possible hunks.
        diff = self.g.diff(oldrev, newrev, '-U0')

        # Get all the hunks.
        hunks = []
        for line in diff.split('\n'):
            if not line.startswith('@@'):
                continue
            ranges = line.split(' ', 3)[1:3]
            ranges = tuple(parse_start_length(r) for r in ranges)
            hunks.append(ranges)

        self.diff_hunks_cache[(oldrev, newrev)] = hunks
        return hunks

    def build_result(self, parsedblame):
        table = []
        for line in parsedblame:
            offset = line.commit.author_tz
            hours = int(offset[:-2])
            minutes = int(offset[-2:])
            tz = timezone(timedelta(hours=hours, minutes=minutes))
            author_time = datetime.utcfromtimestamp(int(
                line.commit.author_time)) + timedelta(hours=hours,
                                                      minutes=minutes)
            author_time = author_time.replace(tzinfo=tz)
            row = ''
            row = [line.commit.commithash[:8],
                   '(' + line.commit.author,
                   author_time.strftime('%Y-%m-%d %H:%M:%S %z'),
                   str(line.lineno_now) + ('*' if line.modified else '') + ')',
                   line.context]
            row.insert(1, line.commit.filename)
            row = ' '.join(row)
            table.append(row)
        return table



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pydriller.domain.commit import Modification, ModificationType

logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)

import pytest
from pathlib import Path
from pydriller.git_repository import GitRepository
from datetime import timezone

@pytest.yield_fixture(scope="module")
def resource():
    yield GitRepository('test-repos/git-1/')


def test_equal(resource):
    c1 = resource.get_commit('e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2')
    c2 = resource.get_commit(c1.parents[0])
    c3 = resource.get_commit('a4ece0762e797d2e2dcbd471115108dd6e05ff58')

    assert c1.parents[0] == 'a4ece0762e797d2e2dcbd471115108dd6e05ff58'
    assert c3 == c2
    assert c1 != c3


def test_filename():
    diff_and_sc = {
        'diff': '',
        'source_code': '',
        'source_code_before': ''
    }
    m1 = Modification('dspadini/pydriller/myfile.py',
                      'dspadini/pydriller/mynewfile.py',
                      ModificationType.ADD, diff_and_sc)
    m3 = Modification('dspadini/pydriller/myfile.py',
                      'dspadini/pydriller/mynewfile.py',
                      ModificationType.ADD, diff_and_sc)
    m2 = Modification('dspadini/pydriller/myfile.py',
                      None,
                      ModificationType.ADD, diff_and_sc)

    assert m1.filename == 'mynewfile.py'
    assert m2.filename == 'myfile.py'
    assert m1 != m2
    assert m3 == m1


def test_metrics_python():
    with open('test-repos/test6/git_repository.py') as f:
        sc = f.read()

    diff_and_sc = {
        'diff': '',
        'source_code': sc,
        'source_code_before': sc
    }

    m1 = Modification('test-repos/test6/git_repository.py',
                      "test-repos/test6/git_repository.py",
                      ModificationType.MODIFY, diff_and_sc)

    assert m1.nloc == 196
    assert m1.token_count == 1009
    assert m1.complexity == 43

    assert len(m1.methods) == 19


def test_metrics_cpp():
    with open('test-repos/test6/FileCPP.cpp') as f:
        sc = f.read()

    diff_and_sc = {
        'diff': '',
        'source_code': sc,
        'source_code_before': sc
    }

    m1 = Modification('test-repos/test6/FileCPP.cpp',
                      "test-repos/test6/FileCPP.cpp",
                      ModificationType.MODIFY, diff_and_sc)

    assert m1.nloc == 793
    assert m1.token_count == 5564
    assert m1.complexity == 199

    assert len(m1.methods) == 16


def test_metrics_java():
    with open('test-repos/test6/FileJava.java') as f:
        sc = f.read()

    diff_and_sc = {
        'diff': '',
        'source_code': sc,
        'source_code_before': sc
    }

    m1 = Modification('test-repos/test6/FileJava.java',
                      "test-repos/test6/FileJava.java",
                      ModificationType.MODIFY, diff_and_sc)

    assert m1.nloc == 466
    assert m1.token_count == 3809
    assert m1.complexity == 92

    assert len(m1.methods) == 46


def test_metrics_not_supported_file():
    sc = 'asd !&%@*&^@\n jjdkj'

    diff_and_sc = {
        'diff': '',
        'source_code': sc,
        'source_code_before': sc
    }

    m1 = Modification('test-repos/test6/NotSupported.pdf',
                      "test-repos/test6/NotSupported.pdf",
                      ModificationType.MODIFY, diff_and_sc)

    assert m1.nloc == 2
    assert len(m1.methods) == 0


def test_filepahs():
    gr = GitRepository('test-repos/test7')
    c = gr.get_commit('f0f8aea2db50ed9f16332d86af3629ff7780583e')

    mod0 = c.modifications[0]

    assert mod0.filename == 'a.java'
    assert mod0.new_path == str(Path('dir2/a.java'))
    assert mod0.old_path == str(Path('dir2/a.java'))


def test_projectname():
    gr = GitRepository('test-repos/test7')
    c = gr.get_commit('f0f8aea2db50ed9f16332d86af3629ff7780583e')

    assert c.project_name == 'test7'


def test_modification_type_unknown():
    gr = GitRepository('test-repos/git-11')
    c = gr.get_commit('1734d6da01378bad3aade12b52bb4aa8954835dc')

    mod0 = c.modifications[0]

    assert mod0.change_type.name == 'UNKNOWN'

    
def test_modification_with_more_parents():
    gr = GitRepository('test-repos/test11')
    c = gr.get_commit('ce6bcd987a6a53cc55da7cef9f8bb128adf68741')
    assert len(c.modifications) == 0

    c = gr.get_commit('1b03d13c816f576eb82a8c3e935fbcacff6c2e8d')
    assert len(c.modifications) == 0


def test_eq_commit():
    gr = GitRepository('test-repos/git-11')
    c1 = gr.get_commit('1734d6da01378bad3aade12b52bb4aa8954835dc')
    c2 = gr.get_commit('2c1327f957ba3b2a5e86eaed097b0a425236719e')
    c3 = gr.get_commit('1734d6da01378bad3aade12b52bb4aa8954835dc')
    m1 = gr.get_commit('1734d6da01378bad3aade12b52bb4aa8954835dc'
                       '').modifications[0]
    assert c1 == c3
    assert c1 == c1
    assert c1 != m1
    assert c1 != c2


def test_eq_modifications():
    gr = GitRepository('test-repos/git-1')
    m1 = gr.get_commit('e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2'
                       '').modifications[0]
    m2 = gr.get_commit('e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2'
                       '').modifications[0]
    m3 = gr.get_commit('a4ece0762e797d2e2dcbd471115108dd6e05ff58'
                       '').modifications[0]
    c1 = gr.get_commit('a4ece0762e797d2e2dcbd471115108dd6e05ff58')

    assert m1 == m2
    assert m1 == m1
    assert m1 != m3
    assert m1 != c1


def test_tzoffset():
    gr = GitRepository('test-repos/git-1')
    tz1 = gr.get_commit(
        'e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2').author_timezone
    tz2 = gr.get_commit(
        'e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2').committer_timezone
    assert tz1 == 10800 # -3 hours
    assert tz2 == 10800 # -3 hours

    gr = GitRepository('test-repos/test1')
    tz1 = gr.get_commit(
        'da39b1326dbc2edfe518b90672734a08f3c13458').author_timezone
    tz2 = gr.get_commit(
        'da39b1326dbc2edfe518b90672734a08f3c13458').committer_timezone
    assert tz1 == -7200 # +2 hours
    assert tz2 == -7200 # +2 hours


def test_source_code_before():
    gr = GitRepository('test-repos/git-1')
    m1 = gr.get_commit('ffccf1e7497eb8136fd66ed5e42bef29677c4b71'
                       '').modifications[0]

    assert m1.source_code is None
    assert m1.source_code_before is not None


def test_source_code_before_complete():
    gr = GitRepository('test-repos/test12')
    m1 = gr.get_commit('ca1f75455f064410360bc56218d0418221cf9484'
                       '').modifications[0]

    with open('test-repos/test12/sc_A_ca1f75455f064410360bc56218d0418221cf9484'
              '.txt') as f:
        sc = f.read()

    assert m1.source_code == sc
    assert m1.source_code_before is None

    old_sc = sc
    with open(
            'test-repos/test12/sc_A_022ebf5fba835c6d95e99eaccc2d85b3db5a2ec0'
            '.txt') as f:
        sc = f.read()

    m1 = gr.get_commit('022ebf5fba835c6d95e99eaccc2d85b3db5a2ec0'
                       '').modifications[0]

    assert m1.source_code == sc
    assert m1.source_code_before == old_sc

    old_sc = sc
    m1 = gr.get_commit('ecd6780457835a2fc85c532338a29f2c98a6cfeb'
                       '').modifications[0]

    assert m1.source_code is None
    assert m1.source_code_before == old_sc
